import calendar
from datetime import datetime
import re
import time

import aniso8601
from flask import url_for, current_app
import six
from werkzeug.utils import cached_property

from . import natural_keys
from .utils import get_value
from .reference import ResourceReference, ResourceBound
from .schema import Schema


class Raw(Schema):
    """
    :param io: one of "r", "w" or "rw" (default); used to control presence in fieldsets/parent schemas
    :param schema: JSON-schema for field, or :class:`callable` resolving to a JSON-schema when called
    :param default: optional default value, must be JSON-convertible
    :param attribute: key on parent object, optional.
    :param nullable: whether the field is nullable.
    :param title: optional title for JSON schema
    :param description: optional description for JSON schema
    """

    def __init__(self, schema, io="rw", default=None, attribute=None, nullable=False, title=None, description=None):
        self._schema = schema
        self.default = default
        self.attribute = attribute
        self.nullable = nullable
        self.title = title
        self.description = description
        self.io = io

    def _finalize_schema(self, schema, io):
        """
        :return: new schema updated for field `nullable`, `title`, `description` and `default` attributes.
        """
        schema = dict(schema)
        
        if self.io == "r" and "r" in io:
            schema["readOnly"] = True

        if "null" in schema.get("type", []):
            self.nullable = True
        elif self.nullable:
            if "anyOf" in schema:
                if not any("null" in choice.get("type", []) for choice in schema["anyOf"]):
                    schema["anyOf"].append({"type": "null"})
            elif "oneOf" in schema:
                if not any("null" in choice.get("type", []) for choice in schema["oneOf"]):
                    schema["oneOf"].append({"type": "null"})
            else:
                try:
                    type_ = schema["type"]
                    if isinstance(type_, (str, dict)):
                        schema["type"] = [type_, "null"]
                    else:
                        schema["type"].append("null")
                except KeyError:
                    if len(schema) == 1 and "$ref" in schema:
                        schema = {"anyOf": [schema, {"type": "null"}]}
                    else:
                        current_app.logger.warn('{} is nullable but "null" type cannot be added'.format(self))

        for attr in ("default", "title", "description"):
            value = getattr(self, attr)
            if value is not None:
                schema[attr] = value
        return schema

    def schema(self):
        """
        JSON schema representation
        """
        schema = self._schema
        if callable(schema):
            schema = schema()

        if isinstance(schema, Schema):
            read_schema, write_schema = schema.response, schema.request
        elif isinstance(schema, tuple):
            read_schema, write_schema = schema
        else:
            return self._finalize_schema(schema, "rw")

        return self._finalize_schema(read_schema, "r"), self._finalize_schema(write_schema, "w")

    def format(self, value):
        """
        Format a Python value representation for output in JSON. Noop by default.
        """
        if value is not None:
            return self.formatter(value)
        return value

    def convert(self, instance, validate=True):
        """
        Convert a JSON value representation to a Python object. Noop by default.
        """
        if validate:
            instance = super(Raw, self).convert(instance)

        if instance is not None:
            return self.converter(instance)
        return instance

    def formatter(self, value):
        return value

    def converter(self, value):
        return value

    def output(self, key, obj):
        key = key if self.attribute is None else self.attribute
        return self.format(get_value(key, obj, self.default))

    def __repr__(self):
        return '{}(attribute={})'.format(self.__class__.__name__, repr(self.attribute))


class Any(Raw):
    def __init__(self, **kwargs):
        super(Any, self).__init__({"type": ["null", "string", "number", "boolean", "object", "array"]}, **kwargs)

def _field_from_object(parent, cls_or_instance):
    # --- start of Flask-RESTful code ---
    # Copyright (c) 2013, Twilio, Inc.
    # All rights reserved.
    # This code is part of or substantially similar to code in Flask-RESTful and is governed by its
    # license. Please see the LICENSE file in the root of this package.
    if isinstance(cls_or_instance, type):
        container = cls_or_instance()
    else:
        container = cls_or_instance
    if not isinstance(container, Schema):
        raise RuntimeError('{} expected Raw or Schema, but got {}'.format(parent, container.__class__.__name__))
    if not isinstance(container, Raw):
        container = Raw(container)
    # --- end of Flask-RESTful code ---
    return container


class Custom(Raw):
    """
    Arbitrary schema field type with optional formatter/converter transformers.

    :param dict schema: JSON-schema
    :param callable converter: convert function
    :param callable formatter: format function
    """

    def __init__(self, schema, converter=None, formatter=None, **kwargs):
        super(Custom, self).__init__(schema, **kwargs)
        self._converter = converter
        self._formatter = formatter

    def format(self, value):
        if self._formatter is None:
            return value
        return self._formatter(value)

    def converter(self, value):
        if self._converter is None:
            return value
        return self._converter(value)


class Array(Raw):
    """
    A field for an array of a given field type.

    :param Raw cls_or_instance: field class or instance
    """

    def __init__(self, cls_or_instance, min_items=None, max_items=None, **kwargs):
        self.container = container = _field_from_object(self, cls_or_instance)

        schema_properties = [('type', 'array')]
        schema_properties += [(k, v) for k, v in [('minItems', min_items), ('maxItems', max_items)] if v is not None]
        schema = lambda s: dict([('items', s)] + schema_properties)

        super(Array, self).__init__(lambda: (schema(container.response), schema(container.request)), **kwargs)

    def formatter(self, value):
        return [self.container.format(v) for v in value]

    def converter(self, value):
        return [self.container.convert(v) for v in value]


List = Array


class Object(Raw):
    """
    A versatile field for an object, containing either properties all of a single type, properties matching a pattern,
    or named properties matching some fields.

    `Raw.attribute` is not used in pattern properties and additional properties.

    :param properties: field class, instance, or dictionary of {property: field} pairs
    :param str pattern: an optional regular expression that all property keys must match
    :param dict pattern_properties: dictionary of {property: field} pairs
    :param Raw additional_properties: field class or instance
    """

    def __init__(self, properties=None, pattern=None, pattern_properties=None, additional_properties=None, **kwargs):
        self.properties = None
        self.pattern_properties = None
        self.additional_properties = None

        if isinstance(properties, dict):
            self.properties = properties
        elif isinstance(properties, (type, Raw)):
            field = _field_from_object(self, properties)
            if pattern:
                self.pattern_properties = {pattern: field}
            else:
                self.additional_properties = field

        def schema():
            request = {"type": "object"}
            response = {"type": "object"}

            for schema, attr in ((request, "request"), (response, "response")):
                if self.properties:
                    schema["properties"] = {key: getattr(field, attr) for key, field in self.properties.items()}
                if self.pattern_properties:
                    schema["patternProperties"] = {pattern: getattr(field, attr)
                                                   for pattern, field in self.pattern_properties.items()}
                if self.additional_properties:
                    schema["additionalProperties"] = getattr(self.additional_properties, attr)
                else:
                    schema["additionalProperties"] = False

            return response, request

        if self.pattern_properties and (len(self.pattern_properties) > 1 or self.additional_properties):
            raise NotImplementedError("Only one pattern property is currently supported "
                                      "and it cannot be combined with additionalProperties")

        super(Object, self).__init__(schema, **kwargs)

    @cached_property
    def _property_attributes(self):
        return [field.attribute or key for key, field in self.properties.items()]

    def format(self, value):
        output = {}

        if self.properties:
            output = {key: field.format(get_value(field.attribute or key, value, field.default)) for key, field in self.properties.items()}
        else:
            output = {}

        if self.pattern_properties:
            pattern, field = next(iter(self.pattern_properties.items()))

            if not self.additional_properties:
                output.update({key: field.format(get_value(key, value, field.default))
                               for key, value in value.items() if key not in self._property_attributes})
            else:
                raise NotImplementedError()
                # TODO match regular expression
        elif self.additional_properties:
            field = self.additional_properties
            output.update({key: field.format(get_value(key, value, field.default))
                           for key, value in value.items() if key not in self._property_attributes})

        return output

    def converter(self, instance):
        result = {}

        if self.properties:
            result = {field.attribute or key: field.convert(instance.get(key, None)) for key, field in self.properties.items()}

        if self.pattern_properties:
            pattern, field = next(iter(self.pattern_properties.items()))

            if not self.additional_properties:
                result.update({key: field.convert(value)
                               for key, value in instance.items() if key not in result})
            else:
                raise NotImplementedError()
                # TODO match regular expression
        elif self.additional_properties:
            field = self.additional_properties
            result.update({key: field.convert(value) for key, value in instance.items() if key not in result})

        return result


class AttributeMapped(Object):
    """
    Maps property keys from a JSON object to a list of items using `mapping_attribute`. The mapping attribute is the
    name of the attribute where the value of the property key is set on the property values.

    .. seealso::

        :class:`InlineModel` field is typically used with this field in a common SQLAlchemy pattern.

    :param Raw cls_or_instance: field class or instance
    :param str pattern: an optional regular expression that all property keys must match
    :param str mapping_attribute: mapping attribute
    """

    def __init__(self, cls_or_instance, mapping_attribute=None, **kwargs):
        self.mapping_attribute = mapping_attribute
        # TODO reject additional_properties, properties, pattern_properties, pattern
        super(AttributeMapped, self).__init__(additional_properties=cls_or_instance, **kwargs)

    def _set_mapping_attribute(self, obj, value):
        setattr(obj, self.mapping_attribute, value)
        return obj

    def format(self, value):
        return {get_value(self.mapping_attribute, v, None): self.additional_properties.format(v) for v in value}

    def converter(self, value):
        return [self._set_mapping_attribute(self.additional_properties.convert(v), k) for k, v in value.items()]


class String(Raw):
    """
    :param int min_length: minimum length of string
    :param int max_length: maximum length of string
    :param str pattern: regex pattern that the string must match
    :param list enum: list of strings with enumeration
    """

    def __init__(self, min_length=None, max_length=None, pattern=None, enum=None, format=None, **kwargs):
        schema = {"type": "string"}

        for v, k in ((min_length, 'minLength'),
                     (max_length, 'maxLength'),
                     (pattern, 'pattern'),
                     (enum, 'enum'),
                     (format, 'format')):
            if v is not None:
                schema[k] = v

        super(String, self).__init__(schema, **kwargs)

try:
    from datetime import timezone
except ImportError:
    from datetime import tzinfo, timedelta

    class timezone(tzinfo):
        def __init__(self, utcoffset, name=None):
            self._utcoffset = utcoffset
            self._name = name

        def utcoffset(self, dt):
            return self._utcoffset

        def tzname(self, dt):
            return self._name

        def dst(self, dt):
            return timedelta(0)

    timezone.utc = timezone(timedelta(0), 'UTC')


class Date(Raw):
    """
    A field for EJSON-style date-times in the format ``{"$date": MILLISECONDS_SINCE_EPOCH}``
    """

    def __init__(self, **kwargs):
        # TODO is a 'format' required for "date"
        super(Date, self).__init__({
                                       "type": "object",
                                       "properties": {
                                           "$date": {
                                               "type": "integer"
                                           }
                                       },
                                       "additionalProperties": False
                                   }, **kwargs)

    def formatter(self, value):
        return {"$date": int(calendar.timegm(value.utctimetuple()) * 1000)}

    def converter(self, value):
        # TODO support both $dateObj and ISO string formats
        return datetime.fromtimestamp(value["$date"] / 1000, timezone.utc)


class DateString(Raw):
    """
    Only accept ISO8601-formatted date strings.
    """

    def __init__(self, **kwargs):
        # TODO is a 'format' required for "date"
        super(DateString, self).__init__({"type": "string", "format": "date"}, **kwargs)

    def format(self, value):
        return value.strftime('%Y-%m-%d')

    def converter(self, value):
        return aniso8601.parse_date(value)


class DateTimeString(Raw):
    """
    Only accept ISO8601-formatted date-time strings.
    """

    def __init__(self, **kwargs):
        super(DateTimeString, self).__init__({"type": "string", "format": "date-time"}, **kwargs)

    def format(self, value):
        return value.isoformat()

    def converter(self, value):
        # FIXME enforce UTC
        return aniso8601.parse_datetime(value)


class Uri(String):
    def __init__(self, **kwargs):
        super(Uri, self).__init__(format="uri", **kwargs)


class Email(String):
    def __init__(self, **kwargs):
        super(Email, self).__init__(format="email", **kwargs)


class Boolean(Raw):
    def __init__(self, **kwargs):
        super(Boolean, self).__init__({"type": "boolean"}, **kwargs)

    def format(self, value):
        return bool(value)


class Integer(Raw):
    def __init__(self, minimum=None, maximum=None, default=None, **kwargs):
        schema = {"type": "integer"}

        if minimum is not None:
            schema['minimum'] = minimum
        if maximum is not None:
            schema['maximum'] = maximum

        super(Integer, self).__init__(schema, default=default, **kwargs)

    def formatter(self, value):
        return int(value)


class PositiveInteger(Integer):
    """
    A :class:`Integer` field that only accepts integers >=1.
    """

    def __init__(self, maximum=None, **kwargs):
        super(PositiveInteger, self).__init__(minimum=1, maximum=maximum, **kwargs)


class Number(Raw):
    def __init__(self,
                 minimum=None,
                 maximum=None,
                 exclusive_minimum=False,
                 exclusive_maximum=False,
                 **kwargs):

        schema = {"type": "number"}

        if minimum is not None:
            schema['minimum'] = minimum
            if exclusive_minimum:
                schema['exclusiveMinimum'] = True

        if maximum is not None:
            schema['maximum'] = maximum
            if exclusive_maximum:
                schema['exclusiveMaximum'] = True

        super(Number, self).__init__(schema, **kwargs)

    def formatter(self, value):
        return float(value)


class ToOne(Raw, ResourceBound):
    def __init__(self, resource, formatter=natural_keys.RefResolver(), **kwargs):
        self.reference = ResourceReference(resource)
        self._formatter = formatter

        def schema():
            target = self.target
            default_schema = formatter.schema(target)
            response_schema = default_schema

            # TODO support $id rather than $refObj
            # natural_keys = target.meta.natural_keys
            # if natural_keys:
            #     request_schema = {
            #         "anyOf": [default_schema] + [nk.schema(target) for nk in natural_keys]
            #     }
            # else:
            #     request_schema = default_schema
            return default_schema

        super(ToOne, self).__init__(schema, **kwargs)

    @cached_property
    def target(self):
        return self.reference.resolve(self.resource)

    def formatter(self, item):
        return self._formatter.format(self.target, item)

    def converter(self, value):
        return self._formatter.resolve(self.target, value)


class ToMany(Array):
    def __init__(self, resource, **kwargs):
        super(ToMany, self).__init__(ToOne(resource, nullable=False), **kwargs)


class Inline(Raw, ResourceBound):

    def __init__(self, resource, patch_instance=False, **kwargs):
        self.reference = ResourceReference(resource)
        self.patch_instance = patch_instance
        self.target = None

        def schema():
            if self.resource == self.target:
                return {"$ref": "#"}
            # FIXME complete with API prefix
            return {"$ref": self.resource.routes["schema"].rule_factory(self.resource)}

        super(Inline, self).__init__(schema, **kwargs)

    def bind(self, resource):
        super(Inline, self).bind(resource)
        self.target = self.reference.resolve(resource)

    def format(self, item):
        return self.target.schema.format(item)

    def convert(self, item):
        return self.target.schema.convert(item, patch_instance=self.patch_instance)


class ItemType(Raw):
    def __init__(self, resource):
        self.resource = resource
        super(ItemType, self).__init__(lambda: {
            "type": "string",
            "enum": [self.resource.meta.name]
        }, io="r")

    def format(self, value):
        return self.resource.meta.name


class ItemUri(Raw):
    def __init__(self, resource, attribute=None):
        self.resource = resource
        super(ItemUri, self).__init__(lambda: {
            "type": "string",
            "pattern": "^{}\/[^/]+$".format(re.escape(resource.route_prefix))
        }, io="r", attribute=attribute)

    def format(self, value):
        return '{}/{}'.format(self.resource.route_prefix, value)


class sa:
    class InlineModel(Object):
        def __init__(self, properties, model, **kwargs):
            super(sa.InlineModel, self).__init__(properties, **kwargs)
            self.model = model

        def converter(self, instance):
            instance = super(sa.InlineModel, self).converter(instance)
            # obj = EmbeddedJob.complete(super().convert(obj))
            # TODO commit()?
            if instance is not None:
                instance = self.model(**instance)
            return instance



from collections import OrderedDict
import operator
from flask import current_app, make_response, json, jsonify
from jsonschema import ValidationError
from six import wraps
from werkzeug.wrappers import BaseResponse
from .exceptions import PotionException
from .utils import unpack
from .resource import Resource, ModelResource
from . import signals, fields

__version_info__ = (1, 0, 0)
__version__ = '.'.join(map(str, __version_info__))
__all__ = (
    'Api',
    'Resource',
    'ModelResource',
    'signals',
    'fields',
)

class Api(object):
    def __init__(self, app=None, decorators=None, prefix=None, default_per_page=20, max_per_page=100):
        self.app = None
        self.prefix = prefix or ''
        self.decorators = decorators or []
        self.max_per_page = max_per_page
        self.default_per_page = default_per_page
        self.endpoints = set()
        self.resources = {}
        self.views = []

        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.app = app
        app.potion = self

        self._complete_view('/schema',
                            view_func=self.output(self._schema_view),
                            endpoint='schema',
                            methods=['GET'])

        for rule, view, endpoint, methods in self.views:
            self._complete_view(rule, view_func=view, endpoint=endpoint, methods=methods)

        @app.errorhandler(PotionException)
        def handle_invalid_usage(error):
            return error.make_response()

    def output(self, view):
        @wraps(view)
        def wrapper(*args, **kwargs):
            resp = view(*args, **kwargs)
            if isinstance(resp, BaseResponse):
                return resp

            data, code, headers = unpack(resp)

            settings = {}
            if current_app.debug:
                settings.setdefault('indent', 4)
                settings.setdefault('sort_keys', True)

            data = json.dumps(data, **settings)
            resp = make_response(data, code)
            resp.headers['Content-Type'] = 'application/json'
            resp.headers.extend(headers or {})
            return resp

        return wrapper

    def _complete_rule(self, rule):
        return ''.join((self.prefix, rule))

    def _complete_view(self, rule, **kwargs):
        self.app.add_url_rule(self._complete_rule(rule), **kwargs)

    def _schema_view(self):
        definitions = OrderedDict([])
        properties = OrderedDict([])
        schema = OrderedDict([
            ("$schema", "http://json-schema.org/draft-04/hyper-schema#"),
            ("definitions", definitions),
            ("properties", properties)
        ])

        # TODO add title, description

        for name, resource in sorted(self.resources.items(), key=operator.itemgetter(0)):
            resource_schema_rule = resource.routes['schema'].rule_factory(resource)
            properties[name] = {"$ref": self._complete_rule('{}#'.format(resource_schema_rule))}

        return schema, 200, {'Content-Type': 'application/schema+json'}

    def add_route(self, route, resource, endpoint=None):
        endpoint = endpoint or '.'.join((resource.meta.name, route.attribute))
        methods = route.methods()
        rule = route.rule_factory(resource)
        view = self.output(route.view_factory(endpoint, resource))

        for decorator in self.decorators:
            view = decorator(view)

        if self.app:
            self._complete_view(rule, view_func=view, endpoint=endpoint, methods=methods)
        else:
            self.views.append((rule, view, endpoint, methods))

    def add_resource(self, resource):
        resource.api = self
        resource.route_prefix = ''.join((self.prefix, '/', resource.meta.name))

        if resource in self.resources.values():
            return

        for route in resource.routes.values():
            self.add_route(route, resource)

        self.resources[resource.meta.name] = resource



from operator import and_
from flask import current_app
from flask_sqlalchemy import get_state
from sqlalchemy import String, func
from sqlalchemy.dialects import postgresql
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import class_mapper
from sqlalchemy.orm.collections import InstrumentedList
from sqlalchemy.orm.exc import NoResultFound
from .. import fields
from ..exceptions import DuplicateKey, ItemNotFound, BackendConflict
from . import Manager, Pagination
from ..signals import before_create, before_update, after_update, before_delete, after_delete, after_create, \
    before_add_to_relation, after_remove_from_relation, before_remove_from_relation, after_add_to_relation

SA_COMPARATOR_EXPRESSIONS = {
    '$eq': lambda column, value: column == value,
    '$ne': lambda column, value: column != value,
    '$in': lambda column, value: column.in_(value) if len(value) else False,
    '$lt': lambda column, value: column < value,
    '$gt': lambda column, value: column > value,
    '$lte': lambda column, value: column <= value,
    '$gte': lambda column, value: column >= value,
    '$text': lambda column, value: column.op('@@')(func.plainto_tsquery(value)),
    '$startswith': lambda column, value: column.startswith(value.replace('%', '\\%')),
    '$endswith': lambda column, value: column.endswith(value.replace('%', '\\%'))
}


class SQLAlchemyManager(Manager):
    supported_comparators = tuple(SA_COMPARATOR_EXPRESSIONS.keys())

    def __init__(self, resource, model):
        super(SQLAlchemyManager, self).__init__(resource, model)

        meta = resource.meta
        mapper = class_mapper(model)

        self.id_attribute = meta.get('id_attribute', mapper.primary_key[0].name)

        if 'id_field' in resource.meta:
            self.id_column = getattr(model, resource.meta.id_field)
        else:
            self.id_column = mapper.primary_key[0]

        # resource name: use model table's name if not set explicitly
        if not hasattr(resource.Meta, 'name'):
            meta['name'] = model.__tablename__.lower()

        fs = resource.schema
        include_fields = meta.get('include_fields', None)
        exclude_fields = meta.get('exclude_fields', None)
        read_only_fields = meta.get('read_only_fields', ())
        write_only_fields = meta.get('write_only_fields', ())
        pre_declared_fields = {f.attribute or k for k, f in fs.fields.items()}

        for name, column in mapper.columns.items():
            if (include_fields and name in include_fields) or \
                    (exclude_fields and name not in exclude_fields) or \
                    not (include_fields or exclude_fields):
                if column.primary_key or column.foreign_keys:
                    continue
                if name in pre_declared_fields:
                    continue

                args = ()
                kwargs = {}

                if isinstance(column.type, postgresql.ARRAY):
                    field_class = fields.Array
                    args = (fields.String,)
                elif isinstance(column.type, String) and column.type.length:
                    field_class = fields.String
                    kwargs = {'max_length': column.type.length}
                elif isinstance(column.type, postgresql.HSTORE):
                    field_class = fields.Object
                    args = (fields.String,)
                elif hasattr(postgresql, 'JSON') and isinstance(column.type, postgresql.JSON):
                    field_class = fields.Raw
                    kwargs = {"schema": {}}
                else:
                    field_class = self._get_field_from_python_type(column.type.python_type)

                kwargs['nullable'] = column.nullable

                if column.default is not None and column.default.is_scalar:
                    kwargs['default'] = column.default.arg

                io = "rw"
                if name in read_only_fields:
                    io = "r"
                elif name in write_only_fields:
                    io = "w"

                # if not (column.nullable or column.default):
                #     fs.required.append(name)
                fs.set(name, field_class(*args, io=io, attribute=name, **kwargs))

    @staticmethod
    def _get_session():
        return get_state(current_app).db.session

    def _query(self):
        return self.model.query

    def _where_expression(self, where):
        expressions = []

        for condition in where:
            column = getattr(self.model, condition.attribute)
            expressions.append(SA_COMPARATOR_EXPRESSIONS[condition.comparator.name](column, condition.value))

        if len(expressions) == 1:
            return expressions[0]

        # TODO ranking by default with text-search.

        return and_(*expressions)

    def _order_by(self, sort):
        for attribute, reverse in sort:
            column = getattr(self.model, attribute)

            if reverse:
                yield column.desc()
            else:
                yield column.asc()

    def relation_instances(self, item, attribute, target_resource, page=None, per_page=None):
        query = getattr(item, attribute)

        if isinstance(query, InstrumentedList):
            if page and per_page:
                return Pagination.from_list(query, page, per_page)
            return query

        if page and per_page:
            return query.paginate(page=page, per_page=per_page)
        return query.all()

    def relation_add(self, item, attribute, target_resource, target_item):
        before_add_to_relation.send(self.resource, item=item, attribute=attribute, child=target_item)
        getattr(item, attribute).append(target_item)
        after_add_to_relation.send(self.resource, item=item, attribute=attribute, child=target_item)

    def relation_remove(self, item, attribute, target_resource, target_item):
        before_remove_from_relation.send(self.resource, item=item, attribute=attribute, child=target_item)
        getattr(item, attribute).remove(target_item)
        after_remove_from_relation.send(self.resource, item=item, attribute=attribute, child=target_item)

    def paginated_instances(self, page, per_page, where=None, sort=None):
        return self.instances(where=where, sort=sort).paginate(page=page, per_page=per_page)

    def instances(self, where=None, sort=None):
        query = self._query()

        if where:
            query = query.filter(self._where_expression(where))
        if sort:
            query = query.order_by(*self._order_by(sort))

        return query

    def create(self, properties, commit=True):
        # noinspection properties
        item = self.model()

        for key, value in properties.items():
            setattr(item, key, value)

        before_create.send(self.resource, item=item)

        session = self._get_session()

        try:
            session.add(item)
            if commit:
                session.commit()
        except IntegrityError as e:
            session.rollback()

            if hasattr(e.orig, 'pgcode'):
                if e.orig.pgcode == "23505":  # duplicate key
                    raise DuplicateKey(detail=e.orig.diag.message_detail)

            if current_app.debug:
                raise BackendConflict(debug_info=dict(statement=e.statement, params=e.params))
            raise BackendConflict()

        after_create.send(self.resource, item=item)
        return item

    def read(self, id):
        try:
            # NOTE SQLAlchemy's .get() does not work well with .filter(), therefore using .one()
            return self._query().filter(self.id_column == id).one()
        except NoResultFound:
            raise ItemNotFound(self.resource, id=id)

    def update(self, item, changes, commit=True):
        session = self._get_session()

        try:
            before_update.send(self.resource, item=item, changes=changes)

            for key, value in changes.items():
                setattr(item, key, value)

            if commit:
                session.commit()
        except IntegrityError as e:
            session.rollback()

            # XXX need some better way to detect postgres engine.
            if hasattr(e.orig, 'pgcode'):
                if e.orig.pgcode == '23505':  # duplicate key
                    raise DuplicateKey(detail=e.orig.diag.message_detail)
            raise

        after_update.send(self.resource, item=item, changes=changes)
        return item

    def delete(self, item):
        before_delete.send(self.resource, item=item)

        session = self._get_session()
        session.delete(item)
        session.commit()

        after_delete.send(self.resource, item=item)


class PrincipalsManager(SQLAlchemyManager):
    pass


from __future__ import division
import datetime
from math import ceil
import six
from .. import fields


class Manager(object):
    supported_comparators = ()

    def __init__(self, resource, model):
        self.resource = resource
        self.model = model

    @staticmethod
    def _get_field_from_python_type(python_type):
        try:
            return {
                str: fields.String,
                six.text_type: fields.String,
                int: fields.Integer,
                float: fields.Number,
                bool: fields.Boolean,
                list: fields.Array,
                dict: fields.Object,
                datetime.date: fields.DateString,
                datetime.datetime: fields.DateTimeString
            }[python_type]
        except KeyError:
            raise RuntimeError('No appropriate field class for "{}" type found'.format(python_type))

    def relation_instances(self, item, attribute, target_resource, page=None, per_page=None):
        raise NotImplementedError()

    def relation_add(self, item, attribute, target_resource, target_item):
        raise NotImplementedError()

    def relation_remove(self, item, attribute, target_resource, target_item):
        raise NotImplementedError()

    def paginated_instances(self, page, per_page, where=None, sort=None):
        pass

    def instances(self, where=None, sort=None):
        pass

    def create(self, properties, commit=True):
        pass

    def read(self, id):
        pass

    def update(self, item, changes, commit=True):
        pass

    def delete(self, item):
        pass

    def delete_by_id(self, id):
        return self.delete(self.read(id))

    def commit(self):
        pass

    def begin(self):
        pass


class Pagination(object):

    def __init__(self, items, page, per_page, total):
        self.items = items
        self.page = page
        self.per_page = per_page
        self.total = total

    @property
    def pages(self):
        return int(ceil(self.total / self.per_page))

    @property
    def has_prev(self):
        return self.page > 1

    @property
    def has_next(self):
        return self.page < self.pages

    @classmethod
    def from_list(cls, items, page, per_page):
        start = per_page * (page - 1)
        return Pagination(items[start:start + per_page], page, per_page, len(items))


from collections import OrderedDict
import inspect
from operator import attrgetter
import itertools
import six
from flask_potion import fields
from .reference import ResourceBound
from .instances import Instances
from .utils import AttributeDict
from .backends.alchemy import SQLAlchemyManager
from .routes import Route, DeferredSchema, RouteSet
from .schema import FieldSet


class ResourceMeta(type):
    def __new__(mcs, name, bases, members):
        class_ = super(ResourceMeta, mcs).__new__(mcs, name, bases, members)
        class_.routes = routes = dict(getattr(class_, 'routes') or {})
        class_.meta = meta = AttributeDict(getattr(class_, 'meta', {}) or {})

        for base in bases:
            for n, m in inspect.getmembers(base, lambda m: isinstance(m, (Route, RouteSet))):
                if m.attribute is None:
                    m.attribute = n

                if isinstance(m, RouteSet):
                    for i, r in enumerate(m.routes()):
                        if r.attribute is None:
                            r.attribute = '{}_{}'.format(m.attribute, i)
                        routes[r.attribute] = r
                else:
                    routes[m.attribute] = m

            if hasattr(base, 'Meta'):
                meta.update(base.Meta.__dict__)


        if 'Meta' in members:
            changes = members['Meta'].__dict__
            for k, v in changes.items():
                if not k.startswith('__'):
                    meta[k] = v

            if not changes.get('name', None):
                meta['name'] = name.lower()

            # NKs: group by type -- e.g. string, integer, object, array -- while keeping order intact:
            # if 'natural_keys' in meta:
            #     meta['natural_keys_by_type'] = natural_keys_by_type = defaultdict(list)
            #     for nk in meta['natural_keys']:
            #         natural_keys_by_type[nk.matcher_type(class_)].append(nk)
        else:
            meta['name'] = name.lower()

        schema = {}
        for base in bases:
            if hasattr(base, 'Schema'):
                schema.update(base.Schema.__dict__)

        if 'Schema' in members:
            schema.update(members['Schema'].__dict__)

        if schema:
            # TODO support FieldSet with definitions
            class_.schema = fs = FieldSet({k: f for k, f in schema.items() if not k.startswith('__')},
                                          required_fields=meta.get('required_fields', None))

            for name in meta.get('read_only_fields', ()):
                if name in fs.fields:
                    fs.fields[name].io = "r"

            for name in meta.get('write_only_fields', ()):
                if name in fs.fields:
                    fs.fields[name].io = "w"

            fs.bind(class_)

        for n, m in members.items():
            if isinstance(m, Route):
                if m.attribute is None:
                    m.attribute = n

                routes[m.attribute] = m

            if isinstance(m, ResourceBound):
                m.bind(class_)

            if isinstance(m, RouteSet):
                if m.attribute is None:
                    m.attribute = n
                for i, r in enumerate(m.routes()):
                    if r.attribute is None:
                        r.attribute = '{}_{}'.format(m.attribute, i)
                    routes['{}_{}'.format(m.attribute, r.attribute)] = r
        return class_


class Resource(six.with_metaclass(ResourceMeta, object)):
    api = None
    meta = None
    routes = None
    schema = None
    route_prefix = None

    @Route.GET('/schema', rel="describedBy", attribute="schema")
    def described_by(self): # No "targetSchema" because that would be way too meta.
        schema = OrderedDict([
            ("$schema", "http://json-schema.org/draft-04/hyper-schema#"),
        ])

        # copy title, description from ModelResource.meta
        for property in ('title', 'description'):
            value = getattr(self.meta, property)
            if value:
                schema[property] = value

        links = itertools.chain(*(route.links() for name, route in sorted(self.routes.items())))

        if self.schema:
            schema['type'] = "object"
            schema.update(self.schema.response)

        # TODO more intuitive sorting [self, instances,.. GET_bar, POST_bar, GET_foo, POST_foo, ..describedBy]
        schema['links'] = [link.schema_factory(self) for link in sorted(links, key=attrgetter('relation'))]

        return schema, 200, {'Content-Type': 'application/schema+json'}

    class Meta:
        name = None
        title = None
        description = None
        required_fields = None
        read_only_fields = ()
        write_only_fields = ()
        # get_item_url_keys = (
        #     RefResolver(),
        # )


class ModelResourceMeta(ResourceMeta):
    def __new__(mcs, name, bases, members):
        class_ = super(ModelResourceMeta, mcs).__new__(mcs, name, bases, members)

        if 'Meta' in members:
            meta = class_.meta
            changes = members['Meta'].__dict__

            if 'model' in changes or 'model' in meta and 'manager' in changes:
                fs = class_.schema

                if issubclass(meta.id_field_class, fields.ItemUri):
                    fs.set('$uri', meta.id_field_class(class_, attribute=meta.id_attribute))
                else:
                    fs.set('$id', meta.id_field_class(io="r", attribute=meta.id_attribute))
                class_.manager = meta.manager(class_, meta.model)

                if meta.include_type:
                    fs.set('$type', fields.ItemType(class_))

        return class_


class ModelResource(six.with_metaclass(ModelResourceMeta, Resource)):
    manager = None

    @Route.GET('', rel="instances")
    def instances(self, **kwargs):
        """
        :param where:
        :param sort:
        :param int page:
        :param int per_page:
        :return:
        """
        return self.manager.paginated_instances(**kwargs)

    # TODO custom schema (Instances/Instances) that contains the necessary schema.
    instances.request_schema = instances.response_schema = DeferredSchema(Instances, 'self') # TODO NOTE Instances('self') for filter, etc. schema


    @instances.POST(rel="create")
    def create(self, properties):  # XXX need some way for field bindings to be dynamic/work dynamically.
        item = self.manager.create(properties)
        return item  # TODO consider 201 Created

    create.request_schema = create.response_schema = DeferredSchema(fields.Inline, 'self')

    @Route.GET(lambda r: '/<{}:id>'.format(r.meta.id_converter), rel="self", attribute="instance")
    def read(self, id):
        return self.manager.read(id)

    read.request_schema = None
    read.response_schema = DeferredSchema(fields.Inline, 'self')

    @read.PATCH(rel="update")
    def update(self, properties, id):
        item = self.manager.read(id)
        updated_item = self.manager.update(item, properties)
        return updated_item

    update.request_schema = DeferredSchema(fields.Inline, 'self', patch_instance=True)
    update.response_schema = update.request_schema

    @update.DELETE(rel="destroy")
    def destroy(self, id):
        self.manager.delete_by_id(id)
        return None, 204

    class Schema:
        pass

    class Meta:
        id_attribute = 'id'
        id_converter = 'int'
        id_field_class = fields.ItemUri  # Must inherit from Integer, String or ItemUri
        include_type = False
        manager = SQLAlchemyManager
        include_fields = None
        exclude_fields = None
        allowed_filters = "*"
        permissions = {
            "read": "anyone",  # anybody, yes
            "create": "nobody",  # noone, no
            "update": "create",
            "delete": "update"
        }
        postgres_text_search_fields = ()
        postgres_full_text_index = None  # $fulltext
        cache = False



from __future__ import division
from ..exceptions import ItemNotFound
from . import Manager, Pagination
from ..signals import before_add_to_relation, after_add_to_relation, before_remove_from_relation, \
    after_remove_from_relation
from ..utils import get_value


class MemoryManager(Manager):
    supported_comparators = ('$eq', '$ne', '$lt', '$gt', '$le', '$ge', '$in', '$startswith', '$endswith')

    def __init__(self, resource, model):
        super(MemoryManager, self).__init__(resource, model)
        self.id_attribute = resource.meta.get('id_attribute', 'id')
        self.id_sequence = 0
        self.items = {}
        self.session = []

    def _new_item_id(self):
        self.id_sequence += 1
        return self.id_sequence

    @staticmethod
    def _filter_items(items, where):
        for item in items:
            # TODO ensure condition uses the field attribute, not the post attribute
            if all(condition(item) for condition in where):
                yield item

    @staticmethod
    def _sort_items(items, sort):
        for key, reverse in reversed(sort):
            items = sorted(items, key=lambda item: get_value(key, item, None), reverse=reverse)
        return items

    def _paginate(self, items, page, per_page):
        return Pagination.from_list(list(items), page, per_page)

    def relation_instances(self, item, attribute, target_resource, page=None, per_page=None):
        collection = item.get(attribute, set())

        items = []
        for id in collection:
            try:
                items.append(target_resource.manager.read(id))
            except ItemNotFound:
                collection.remove(id)
                pass

        return Pagination.from_list(items, page, per_page)

    def relation_add(self, item, attribute, target_resource, target_item):
        before_add_to_relation.send(self.resource, item=item, attribute=attribute, child=target_item)
        item[attribute] = collection = item.get(attribute, set())
        item_id = target_item[target_resource.manager.id_attribute]
        collection.add(item_id)
        after_add_to_relation.send(self.resource, item=item, attribute=attribute, child=target_item)


    def relation_remove(self, item, attribute, target_resource, target_item):
        before_remove_from_relation.send(self.resource, item=item, attribute=attribute, child=target_item)
        item[attribute] = collection = item.get(attribute, set())
        item_id = target_item[target_resource.manager.id_attribute]
        collection.remove(item_id)
        after_remove_from_relation.send(self.resource, item=item, attribute=attribute, child=target_item)

    def paginated_instances(self, page, per_page, where=None, sort=None):
        return self._paginate(self.instances(where=where, sort=sort), page, per_page)

    def instances(self, where=None, sort=None):
        items = self.items.values()

        if where is not None:
            items = self._filter_items(items, where)
        if sort is not None:
            items = self._sort_items(items, sort)

        return items

    def create(self, properties, commit=True):
        item_id = self._new_item_id()
        item = dict({self.id_attribute: item_id})
        item.update(properties)

        if commit:
            self.items[item_id] = item
        else:
            self.session[item_id] = item

        return item

    def read(self, id):
        try:
            item = self.items[id]
        except KeyError:
            raise ItemNotFound(self.resource, id=id)

        return item

    def update(self, item, changes, commit=True):
        item_id = item[self.id_attribute]
        item = dict(item)
        item.update(changes)

        if commit:
            self.items[item_id] = item
        else:
            self.session.append((item_id, item))

        return item

    def delete(self, item):
        item_id = item[self.id_attribute]
        del self.items[item_id]

    def commit(self):
        for item_id, item in self.session:
            self.items[item_id] = item

    def begin(self):
        self.session = []


from collections import OrderedDict
import re
from types import MethodType

from flask import request
import sys
from werkzeug.utils import cached_property
from flask_potion.fields import _field_from_object
from flask_potion import fields
from .utils import get_value
from .instances import Instances, RelationInstances
from .reference import ResourceBound, ResourceReference
from .schema import Schema, FieldSet


HTTP_METHODS = ('GET', 'PUT', 'POST', 'PATCH', 'DELETE')

def url_rule_to_uri_pattern(rule):
    # TODO convert from underscore to camelCase
    return re.sub(r'<(\w+:)?([^>]+)>', r'{\2}', rule)

def attribute_to_route_uri(s):
    return s.replace('_', '-')

def to_camel_case(s):
    return s[0].lower() + s.title().replace('_', '')[1:] if s else s

class DeferredSchema(object):
    def __init__(self, class_, *args, **kwargs):
        self.schema_class = class_
        self.schema_args = args
        self.schema_kwargs = kwargs
        self._cached_schemas = {}

    def __call__(self, resource):
        if resource in self._cached_schemas:
            return self._cached_schemas[resource]

        schema = self.schema_class(*self.schema_args, **self.schema_kwargs)
        if isinstance(schema, ResourceBound):
            schema.bind(resource)

        self._cached_schemas[resource] = schema
        return schema

    @classmethod
    def resolve(cls, schema, resource):
        if isinstance(schema, cls):
            return schema(resource)
        if isinstance(schema, ResourceBound):
            schema.bind(resource)
        return schema


class Link(object):
    def __init__(self,
                 view_func,
                 rel=None,
                 route=None,
                 method=None,
                 schema=None,
                 response_schema=None,
                 format_response=True):
        self.rel = rel
        self.route = route
        self.method = method
        self.view_func = view_func
        self.format_response = format_response

        annotations = getattr(view_func, '__annotations__', None)

        if isinstance(annotations, dict) and len(annotations):
            self.request_schema = FieldSet({name: field for name, field in annotations.items() if name != 'return'})
            self.response_schema = annotations.get('return', response_schema)
        else:
            self.request_schema = schema
            self.response_schema = response_schema

    @property
    def relation(self):
        if self.rel:
            return self.rel
        elif len(self.route.methods()) == 1:
            return self.route.attribute
        else:
            return "{}_{}".format(self.method, self.route.attribute)

    def schema_factory(self, resource):
        request_schema = DeferredSchema.resolve(self.request_schema, resource)
        response_schema = DeferredSchema.resolve(self.response_schema, resource)

        # NOTE "href" for rel="instances" MUST NOT be relative; others MAY BE relative
        schema = OrderedDict([
            ("rel", self.relation),
            ("href", url_rule_to_uri_pattern(self.route.rule_factory(resource, relative=False))),
            ("method", self.method)
        ])

        if request_schema:
            schema["schema"] = request_schema.request
        if response_schema:
            schema["targetSchema"] = response_schema.response
        return schema


def _method_decorator(method):
    def wrapper(self, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return self._set_method_link(method, args[0], **kwargs)
        else:
            return lambda f: self._set_method_link(method, f, *args, **kwargs)

    wrapper.__name__ = method
    return wrapper


class Route(object):
    def __init__(self, rule=None, attribute=None, format_response=True):
        self.rule = rule
        self.attribute = attribute
        self.format_response = format_response
        self.current_link = None
        self._method_links = {}

        for method in HTTP_METHODS:
            setattr(self, method, MethodType(_method_decorator(method), self))

    @classmethod
    def for_method(cls, method, func, rule=None, **kwargs):
        instance = cls(rule=rule,
                       attribute=kwargs.pop('attribute', None),
                       format_response=kwargs.pop('format_response', True))
        instance._set_method_link(method, func, **kwargs)
        return instance

    def _set_method_link(self, method, view_func, rel=None, schema=None, response_schema=None):
        link = Link(view_func,
                        rel=rel,
                        route=self,
                        method=method,
                        schema=schema,
                        response_schema=response_schema)
        self.current_link = link
        self._method_links[method] = link
        return self

    def __getattr__(self, name):
        return getattr(self.current_link, name)

    def __get__(self, obj, owner):
        if obj is None:
            return self
        return lambda *args, **kwargs: self.current_link.view_func.__call__(obj, *args, **kwargs)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self.rule))

    @property
    def request_schema(self):
        return self.current_link.request_schema

    @request_schema.setter
    def request_schema(self, schema):
        self.current_link.request_schema = schema

    @property
    def response_schema(self):
        return self.current_link.response_schema

    @response_schema.setter
    def response_schema(self, schema):
        self.current_link.response_schema = schema

    def links(self):
        return self._method_links.values()

    def methods(self):
        return [link.method for link in self.links()]

    def _rule_factory(self, resource, relative=False):
        rule = self.rule

        if rule is None:
            rule = '/{}'.format(attribute_to_route_uri(self.attribute))
        elif callable(rule):
            rule = rule(resource)

        if relative or resource.route_prefix is None:
            return rule[1:]

        return ''.join((resource.route_prefix, rule))

    def _view_factory(self, link, name, resource):
        request_schema = DeferredSchema.resolve(link.request_schema, resource)
        response_schema = DeferredSchema.resolve(link.response_schema, resource)
        view_func = link.view_func

        def view(*args, **kwargs):
            instance = resource()
            if isinstance(request_schema, (FieldSet, Instances)):
                kwargs.update(request_schema.parse_request(request))
            elif isinstance(request_schema, Schema):
                args += (request_schema.parse_request(request),)

            response = view_func(instance, *args, **kwargs)
            # TODO add 'described_by' header if response schema is a ToOne/ToMany/Instances field.
            if response_schema is None or not self.format_response:
                return response
            else:
                return response_schema.format_response(response)

        return view

    def rule_factory(self, resource, relative=False):
        return self._rule_factory(resource, relative)

    def view_factory(self, name, resource):
        method_views = {}

        for name, link in self._method_links.items():
            method_views[name] = self._view_factory(link, name, resource)

        def view(*args, **kwargs):
            method = method_views.get(request.method)
            if method is None and request.method == 'HEAD':
                method = method_views['GET']
            return method(*args, **kwargs)
        return view


def _route_decorator(method):
    @classmethod
    def decorator(cls, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return cls.for_method(method, args[0])
        else:
            return lambda f: cls.for_method(method, f, *args, **kwargs)
    if sys.version_info.major > 2:
        decorator.__name__ = method
    return decorator

for method in HTTP_METHODS:
    setattr(Route, method, _route_decorator(method))


class ItemRoute(Route):
    def _rule_factory(self, resource, relative=False):
        rule = self.rule
        id_matcher = '<{}:id>'.format(resource.meta.id_converter)

        if rule is None:
            rule = '/{}'.format(attribute_to_route_uri(self.attribute))
        elif callable(rule):
            rule = rule(resource)

        if relative:
            return ''.join((id_matcher, '/', rule[1:]))
        return ''.join(('/', resource.meta.name, '/', id_matcher, rule))

    def _view_factory(self, link, name, resource):
        original_view = super(ItemRoute, self)._view_factory(link, name, resource)
        def view(*args, **kwargs):
            id = kwargs.pop('id')  # Py2.7 -- could use (*args, id, **kwargs) otherwise
            item = resource.manager.read(id)
            return original_view(item, *args, **kwargs)
        return view


class RouteSet(object):
    # TODO consider adding dependencies() function for resolving relations before the routes are added

    def routes(self):
        pass


class ItemAttributeRoute(RouteSet):
    def __init__(self, cls_or_instance, io=None, attribute=None):
        self.field = _field_from_object(ItemAttributeRoute, cls_or_instance)
        self.attribute = attribute
        self.io = io

    def routes(self):
        io = self.io or self.field.io
        field = self.field
        route = ItemRoute(attribute=self.attribute)
        attribute = field.attribute or route.attribute

        if "r" in io:
            @route.GET(response_schema=field,
                       rel=to_camel_case('read_{}'.format(route.attribute)))
            def read_attribute(resource, item):
                return get_value(attribute, item, field.default)

        if "w" in io:
            @route.POST(schema=field,
                        response_schema=field,
                        rel=to_camel_case('update_{}'.format(route.attribute)))
            def update_attribute(resource, item, value):
                attribute = field.attribute or route.attribute
                item = resource.manager.update(item, {attribute: value})
                return get_value(attribute, item, field.default)

        yield route


class Relation(RouteSet, ResourceBound):
    def __init__(self, resource, backref=None, io="rw", attribute=None, **kwargs):
        self.reference = ResourceReference(resource)
        self.attribute = attribute
        self.backref = backref
        self.io = io

    @cached_property
    def target(self):
        return self.reference.resolve(self.resource)

    # FIXME can only be loaded after target is added to API
    def routes(self):
        io = self.io
        rule = '/{}'.format(attribute_to_route_uri(self.attribute))

        relation_route = ItemRoute(rule='{}/<{}:target_id>'.format(rule, self.target.meta.id_converter))
        relations_route = ItemRoute(rule=rule)

        if "r" in io:
            @relations_route.GET
            def relation_instances(resource, item, page, per_page):
                return resource.manager.relation_instances(item,
                                                           self.attribute,
                                                           self.target,
                                                           page,
                                                           per_page)

            relations_route.request_schema = FieldSet({
                "page": fields.Integer(minimum=1, default=1),
                "per_page": fields.Integer(minimum=1,
                                           default=20, # FIXME use API reference
                                           maximum=50)
            })

            relations_route.response_schema = RelationInstances(self.target)

        if "w" in io:
            @relations_route.POST
            def relation_add(resource, item, target_item):
                resource.manager.relation_add(item, self.attribute, self.target, target_item)
                resource.manager.commit()
                return target_item

            relation_add.request_schema = fields.ToOne(self.target)
            relation_add.response_schema = fields.ToOne(self.target)

            @relation_route.DELETE
            def relation_remove(resource, item, target_id):
                target_item = self.target.manager.read(target_id)
                resource.manager.relation_remove(item, self.attribute, self.target, target_item)
                resource.manager.commit()
                return None, 204
            yield relation_route

        if io:
            yield relations_route


class ItemMapAttribute(RouteSet):
    """
    Adds a route that includes the keys to a dictionary attribute and allows these to be read, set, and deleted
    either individually or in bulk.

    class Schema:
        scores = fields.Object(field.Integer)

    GET /item/:id/scores/:key/
    PUT /item/:id/scores/:key/
    DELETE /item/:id/scores/:key/

    :param str mapping_attribute: Must map to an attribute that is of type :class:`str`
    """

    def __init__(self, cls_or_instance, io=None, **kwargs):
        self.field = field = _field_from_object(cls_or_instance)
        self.object_field = fields.AttributeMapped(field, **kwargs)
        self.io = io

    # @Route.GET(lambda r: '/<id:{}>/attribute/<key>'.format(r.meta.id_converter))
    # def attribute(self, item, key):
    #     pass

    def routes(self):
        field = self.field
        object_field = self.object_field
        attribute = field.attribute or self.attribute
        rule = '/{}'.format(attribute_to_route_uri(attribute))
        io = self.io or field.io

        object = ItemRoute(rule=rule)
        # object_property = ItemRoute(rule=rule)

        @object.GET(response_schema=object_field)
        def read(resource, item):
            return get_value(attribute, item, {})

        read.response_schema = object_field

        @object.POST(schema=object_field)
        def write(resource, item, value):
            resource.manager.update(item, {attribute: value})
            return value

        # TODO PATCH object

        # @object_property.GET(response_schema=field)
        # def read_property(resource, item, key, value):
        #     # TODO ensure key matches pattern.
        #     pass
        #
        # @object_property.POST(schema=field)
        # def write_property(resource, item, key, value):
        #     # TODO ensure key matches pattern.
        #     pass
        #
        # @object_property.DELETE
        # def remove_property(resource, item, key):
        #     # TODO ensure key matches pattern.
        #     pass

        yield object


class ItemSetAttributeRoute(ItemRoute):
    """
    GET /item/:id/values
    POST /item/:id/value/:value_id
    DELETE /item/:id/value/:value_id

    The URL of each item must stay consistent when the order of the list changes.

    Direct addressing is only supported when an `id_attribute` is given. This must be an attribute
    that is unique and cannot be changed. Without it, only read-only mode can be supported.

    Depending on the type of collection, sorting and filtering may or may not be supported.

    """

    def __init__(self, cls_or_instance, id_attribute=None, id_field=None, io="r"):
        if "w" in io:
            raise NotImplementedError("ItemListAttributeRoute only supports read-only mode for now.")

        pass

    @ItemRoute.GET()
    def read_items(self):
        pass

    @ItemRoute.GET()
    def read_item(self, item, position):
        pass

    @read_item.POST()
    def add_item(self, item):
        pass



from importlib import import_module
import inspect
import six
from flask import current_app


class ResourceReference(object):
    def __init__(self, value):
        self.value = value

    def resolve(self, binding=None):
        """
        Attempt to resolve the reference value and return the matching :class:`Resource`.
        """
        name = self.value

        if name == 'self':
            return binding

        from .resource import ModelResource
        if inspect.isclass(name) and issubclass(name, ModelResource):
            return name

        # FIXME XXX need a better back-reference to the Potion instance
        if hasattr(current_app, 'potion'):
            if name in current_app.potion.resources:
                return current_app.potion.resources[name]

        try:
            if isinstance(name, six.string_types):
                module_name, class_name = name.rsplit('.', 1)
                module = import_module(module_name)
                return getattr(module, class_name)
        except ValueError:
            pass

        raise RuntimeError('Resource named "{}" is not registered with Potion'.format(name))

    def __repr__(self):
        return "<ResourceReference '{}'>".format(self.value)


class ResourceBound(object):
    resource = None

    def bind(self, resource):
        self.resource = resource


from flask import g
from flask_principal import Permission
from sqlalchemy import or_
from .needs import HybridNeed


class HybridPermission(Permission):
    """
    Hybrid Permission object that evaluates both regular and hybrid needs
    """
    def __init__(self, *needs):
        super(HybridPermission, self).__init__(*needs)
        self.hybrid_needs = set()
        self.standard_needs = set()

        for need in needs:
            if isinstance(need, HybridNeed):
                self.hybrid_needs.add(need)
            else:
                self.standard_needs.add(need)

    def allows(self, identity):
        """Whether the identity can access this permission.

        :param identity: The identity
        """
        if self.standard_needs and not self.standard_needs.intersection(identity.provides):
            return False

        # TODO support standard needs for excludes
        if self.excludes and self.excludes.intersection(identity.provides):
            return False

        if self.needs and not self.standard_needs:
            return False

        return True

    def can(self, item=None):
        """
        Evaluate either only the standard needs, or if ``item`` is given also evaluate the hybrid needs on the item.
        """
        if not item:
            return self.require().can()
        else:
            if self.require().can():
                return True

            for need in self.hybrid_needs:
                resolved_need = need(item)
                if resolved_need in g.identity.provides:
                    return True
        return False

    def apply_filters(self, query):
        """
        Evaluates all *needs* including :class:`HybridNeed` types and filters the query as appropriate.

        Multiple hybrid needs are combined using `or`. That is, only one has to match.

        :returns: `None` if no hybrid needs are present; *query* object otherwise.
        """
        hybrid_relationship_need = None

        # prefer not to filter at all:
        if self.can():
            return query

        # filters must not be applied if not present:
        if not self.hybrid_needs:
            return None

        expressions = []

        for need in self.hybrid_needs:
            expression = need.make_expression()
            if expression is not None:
                expressions.append(expression)

        if not expressions:
            return None
        if len(expressions) == 1:
            return query.filter(expressions.pop())
        else:
            return query.filter(or_(*expressions))



import re
from flask import url_for
from werkzeug.utils import cached_property
from .exceptions import ItemNotFound
from .schema import Schema
from .utils import route_from, get_value
import six

class Resolver(object):

    def matcher_type(self, resource):
        type_ = self.schema(resource)['type']
        if isinstance(type_, six.text_type):
            return type_
        return type_[0]

    def schema(self, resource):
        raise NotImplementedError()


class RefResolver(Resolver):

    def schema(self, resource):
        return {
            "type": "object",
            "properties": {
                "$ref": {
                    "type": "string",
                    "format": "uri",
                    "pattern": "^{}\/[^/]+$".format(re.escape(resource.route_prefix))
                }
                # TODO consider replacing with {$type: foo, $value: 123}
            },
            "additionalProperties": False
        }

    def _item_uri(self, resource, item):
        # return url_for('{}.instance'.format(self.resource.meta.id_attribute, item, None), get_value(self.resource.meta.id_attribute, item, None))
        return '{}/{}'.format(resource.route_prefix, get_value(resource.meta.id_attribute, item, None))

    def format(self, resource, item):
        return {"$ref": self._item_uri(resource, item)}

    def resolve(self, resource, value):
        try:
            endpoint, args = route_from(value["$ref"], 'GET')
        except Exception as e:
            raise e
        # XXX verify endpoint is correct (it should be)
        # assert resource.endpoint == endpoint
        return resource.manager.read(args['id'])


class PropertyResolver(Resolver):

    def __init__(self, property):
        self.property = property

    def schema(self, resource):
        return resource.schema.fields[self.property].response

    def format(self, resource, item):
        return resource.schema.fields[self.property].output(self.property, item)

    def resolve(self, resource, value):
        instances = resource.manager.instances(where={self.property: value})
        try:
            return instances[0]
        except IndexError:
            raise ItemNotFound(resource, natural_key=value)


class PropertiesResolver(Resolver):
    def __init__(self, *properties):
        self.properties = properties

    def schema(self, resource):
        return {
            "type": "array",
            "items": [resource.schema.fields[p].response for p in self.properties],
            "additionalItems": False
        }

    def format(self, resource, item):
        return [resource.schema.fields[p].output(p, item) for p in self.properties]

    def resolve(self, resource, value):
        instances = resource.manager.instances(where={property: value[i] for i, property in enumerate(self.properties)})
        try:
            return instances[0]
        except IndexError:
            raise ItemNotFound(resource, natural_key=value)


class IDResolver(Resolver):
    def schema(self, resource):
        return resource.schema.fields['$id'].response

    def format(self, resource, item):
        return resource.schema.fields['$id'].output(resource.meta.id_attribute, item)

    def resolve(self, resource, value):
        return resource.manager.read(value)



from blinker import Namespace

_potion = Namespace()

before_create = _potion.signal('before-create')

after_create = _potion.signal('after-create')

before_update = _potion.signal('before-update')

after_update = _potion.signal('after-update')

before_delete = _potion.signal('before-delete')

after_delete = _potion.signal('after-delete')

before_add_to_relation = _potion.signal('before-add-to-relation')

after_add_to_relation = _potion.signal('after-add-to-relation')

before_remove_from_relation = _potion.signal('before-remove-from-relation')

after_remove_from_relation = _potion.signal('after-remove-from-relation')


from flask import jsonify
from werkzeug.exceptions import Conflict, BadRequest, NotFound, InternalServerError
from werkzeug.http import HTTP_STATUS_CODES


class PotionException(Exception):
    http_exception = InternalServerError

    @property
    def status_code(self):
        return self.http_exception.code

    def as_dict(self):
        return {
            'status': self.status_code,
            'message': HTTP_STATUS_CODES.get(self.status_code, '')
        }

    def make_response(self):
        code = self.http_exception.code
        response = jsonify(self.as_dict())
        response.status_code = self.status_code
        return response


class ItemNotFound(PotionException):
    http_exception = NotFound

    def __init__(self, resource, natural_key=None, id=None):
        super(ItemNotFound, self).__init__()
        self.resource = resource
        self.id = id
        self.natural_key = natural_key

    def as_dict(self):
        dct = super(ItemNotFound, self).as_dict()

        if self.id is not None:
            dct['item'] = {
                "$type": self.resource.meta.name,
                "$id": self.id
            }
        return dct

    def make_response(self):
        code = self.http_exception.code
        response = jsonify(self.as_dict())
        response.status_code = self.status_code
        return response


class ValidationError(PotionException):
    http_exception = BadRequest

    def __init__(self, errors, root=None, schema_uri='#'):
        self.root = root
        self.errors = errors
        self.schema_uri = schema_uri

    def _complete_path(self, error):
        path = tuple(error.absolute_path)
        if self.root is not None:
            return (self.root, ) + path
        return path

    def as_dict(self):
        dct = super(ValidationError, self).as_dict()

        dct['errors'] = [
            {
                'validationOf': {error.validator: error.validator_value},
                "path": self._complete_path(error)
                # 'message': error.message
            }
            for error in self.errors
        ]

        return dct


class DuplicateKey(PotionException):
    http_exception = Conflict

    def __init__(self, **kwargs):
        self.data = kwargs


class BackendConflict(PotionException):
    http_exception = Conflict

    def __init__(self, **kwargs):
        self.data = kwargs

    def as_dict(self):
        dct = super(BackendConflict, self).as_dict()
        dct.update(self.data)
        return dct

class PageNotFound(PotionException):
    http_exception = NotFound


class InvalidJSON(PotionException):
    http_exception = BadRequest



from flask import g
from flask_principal import UserNeed, ItemNeed
from sqlalchemy.orm.attributes import ScalarObjectAttributeImpl


class HybridNeed(object):
    """
    :class:`HybridNeed` base class. Hybrid needs can both be evaluated directly or produce an expression for use with
    SQLAlchemy.
    """
    def __call__(self, item):
        raise NotImplementedError()

    def __hash__(self):
        return hash(self.__repr__())

    def make_expression(self, query):
        """
        :returns: SQLAlchemy expression for this Need
        """
        # TODO support inversion of HybridNeed objects for negative permissions.
        raise NotImplementedError()


class HybridItemNeed(HybridNeed):
    def __init__(self, method, resource, type_=None):
        self.method = method
        self.type = type_ or resource.meta.name
        self.resource = resource

    def _identity_get_item_needs(self):
        if self.method == 'id':
            prototype = ('id', None)
        else:
            prototype = (self.method, None, self.type)

        for need in g.identity.provides:
            if len(need) == len(prototype):
                if all(p is None or n == p for n, p in zip(need, prototype)):
                    yield need[1]

    def extend(self, field):
        return HybridRelationshipNeed(self.method, field)

    def __call__(self, item):
        if self.method == 'id':
            return UserNeed(self.resource.item_get_id(item))
        return ItemNeed(self.method, get_value(item, self.resource.meta.id_attribute, None), self.type)

    def __eq__(self, other):
        return isinstance(other, HybridItemNeed) and \
               self.method == other.method and \
               self.type == other.type and \
               self.resource == other.resource

    def make_expression(self):
        ids = list(self._identity_get_item_needs())

        if not ids:
            return None

        return self.resource.manager.id_column.in_(ids)

    def __hash__(self):
        return hash(self.__repr__())

    def __repr__(self):
        return "<HybridItemNeed method='{}' type='{}'>".format(self.method, self.type)


def get_value(item, attribute, default=None):
    if isinstance(item, dict):
        return item.get(attribute, default)
    else:
        return getattr(item, attribute, None)


class HybridRelationshipNeed(HybridItemNeed):
    """
    HybridNeed objects
    """

    def __init__(self, method, *fields):
        super(HybridRelationshipNeed, self).__init__(method,
                                                     fields[-1].resource,
                                                     fields[-1].target.meta.name)
        self.fields = fields
        self.final_field = self.fields[-1]

    def __call__(self, item):
        """

        """
        for field in self.fields:
            item = get_value(item, field.attribute)

            if item is None:
                if self.method == 'id':
                    return UserNeed(None)
                return ItemNeed(self.method, None, self.type)

        item_id = get_value(item, self.final_field.resource.meta.id_attribute, None)

        if self.method == 'id':
            return UserNeed(item_id)
        return ItemNeed(self.method, item_id, self.type)

    def __eq__(self, other):
        return isinstance(other, HybridItemNeed) and \
               self.method == other.method and \
               self.resource == other.resource and \
               self.fields == other.fields

    def extend(self, field):
        return HybridRelationshipNeed(self.method, field, *self.fields)

    def make_expression(self):
        ids = list(self._identity_get_item_needs())

        if not ids:
            return None

        reversed_fields = reversed(self.fields)

        target_field = next(reversed_fields)
        target_relationship = getattr(target_field.resource.manager.model, target_field.attribute)

        expression = target_relationship.has(target_field.target.manager.id_column.in_(ids))

        for field in reversed_fields:
            relationship = getattr(field.binding._model, field.attribute)

            if isinstance(relationship.impl, ScalarObjectAttributeImpl):
                expression = relationship.has(expression)
            else:
                expression = relationship.any(expression)

        return expression

    def __hash__(self):
        return hash((self.method, self.type, self.fields))

    def __repr__(self):
        return "<HybridRelationshipNeed method='{}' type='{}' {}>".format(self.method, self.type, self.fields)


class HybridUserNeed(HybridRelationshipNeed):
    def __init__(self, field):
        super(HybridUserNeed, self).__init__('id', field)

    def __repr__(self):
        return '<HybridUserNeed {} {}>'.format(self.type, self.fields)




from functools import wraps
from flask import _app_ctx_stack, _request_ctx_stack
from werkzeug.exceptions import NotFound
from werkzeug.urls import url_parse


def route_from(url, method=None):
    # Source: http://stackoverflow.com/questions/19129407/looking-for-inverse-of-url-for-in-flask
    appctx = _app_ctx_stack.top
    reqctx = _request_ctx_stack.top
    if appctx is None:
        raise RuntimeError('Attempted to match a URL without the '
                           'application context being pushed. This has to be '
                           'executed when application context is available.')

    if reqctx is not None:
        url_adapter = reqctx.url_adapter
    else:
        url_adapter = appctx.url_adapter
        if url_adapter is None:
            raise RuntimeError('Application was not able to create a URL '
                               'adapter for request independent URL matching. '
                               'You might be able to fix this by setting '
                               'the SERVER_NAME config variable.')
    parsed_url = url_parse(url)
    if parsed_url.netloc is not "" and parsed_url.netloc != url_adapter.server_name:
        raise NotFound()
    return url_adapter.match(parsed_url.path, method)


# --- start of Flask-RESTful code ---
# Copyright (c) 2013, Twilio, Inc.
# All rights reserved.
# This code is part of Flask-RESTful and is governed by its
# license. Please see the LICENSE file in the root of this package.
def unpack(value):
    """Return a three tuple of data, code, and headers"""
    if not isinstance(value, tuple):
        return value, 200, {}

    try:
        data, code, headers = value
        return data, code, headers
    except ValueError:
        pass

    try:
        data, code = value
        return data, code, {}
    except ValueError:
        pass

    return value, 200, {}


def get_value(key, obj, default):
    if hasattr(obj, '__getitem__'):
        try:
            return obj[key]
        except (IndexError, TypeError, KeyError):
            pass
    return getattr(obj, key, default)
# --- end of Flask-RESTful code ---


class AttributeDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__



from collections import OrderedDict
from flask import json
from werkzeug.utils import cached_property
from jsonschema import Draft4Validator, ValidationError, FormatChecker
from .reference import ResourceBound
from .utils import unpack
from .exceptions import ValidationError as PotionValidationError


class Schema(object):

    def schema(self):
        raise NotImplementedError()

    @cached_property
    def response(self):
        schema = self.schema()
        if isinstance(schema, tuple):
            return schema[0]
        return schema

    @cached_property
    def request(self):
        schema = self.schema()
        if isinstance(schema, tuple):
            return schema[1]
        return schema

    @cached_property
    def _validator(self):
        Draft4Validator.check_schema(self.request)
        return Draft4Validator(self.request, format_checker=FormatChecker())

    def format(self, value):
        return value

    def convert(self, instance):
        try:
            self._validator.validate(instance)
        except ValidationError as ve:
            errors = self._validator.iter_errors(instance)
            raise PotionValidationError(errors)
        return instance

    def parse_request(self, request):
        data = request.json

        if not data and request.method in ('GET', 'HEAD'):
            data = dict(request.args)

        return self.convert(data)

    def format_response(self, response):
        data, code, headers = unpack(response)
        # TODO omit formatting on certain codes.
        return self.format(data), code, headers


class FieldSet(Schema, ResourceBound):

    def __init__(self, fields, required_fields=None):
        self.fields = fields
        self.required = required_fields or ()

    def bind(self, resource):
        ResourceBound.bind(self, resource)
        for key, field in self.fields.items():
            if isinstance(field, ResourceBound):
                field.bind(resource)

    def set(self, key, field):
        self.fields[key] = field
        if self.resource and isinstance(field, ResourceBound):
            field.bind(self.resource)

    def schema(self):
        response_schema = {
            "type": "object",
            "properties": OrderedDict((
                (key, field.response) for key, field in self.fields.items() if 'r' in field.io))
        }

        request_schema = {
            "type": "object",
            "additionalProperties": False,
            "properties": OrderedDict((
                (key, field.request) for key, field in self.fields.items() if 'w' in field.io))
        }

        if self.required:
            request_schema['required'] = list(self.required)

        return response_schema, request_schema

    def format(self, item):
        return OrderedDict((key, field.output(key, item)) for key, field in self.fields.items())

    def convert(self, instance, pre_resolved_properties=None, patch_instance=False, strict=False):
        result = dict(pre_resolved_properties) if pre_resolved_properties else {}
        object_ = super(FieldSet, self).convert(instance)

        for key, field in self.fields.items():
            if 'w' not in field.io:
                continue

            # ignore fields that have been pre-resolved
            if key in result:
                continue

            value = None

            try:
                value = object_[key]
                value = field.convert(value, validate=False)
            except KeyError:
                if patch_instance:
                    continue

                if field.default is not None:
                    value = field.default
                elif field.nullable:
                    value = None
                elif key not in self.required and not strict:
                    value = None

            result[field.attribute or key] = value
        return result

    def parse_request(self, request):
        data = request.json

        # FIXME raise error if request body is not JSON

        if not self.fields:
            return {}

        if not data and request.method in ('GET', 'HEAD'):
            data = {}

            for name, field in self.fields.items():
                try:
                    value = request.args[name]
                    # FIXME type conversion!
                    try:
                        data[name] = json.loads(value)
                    except TypeError:
                        data[name] = value
                except KeyError:
                    pass

        return self.convert(data, patch_instance=request.method == 'PATCH')




import six
from sqlalchemy.orm.collections import InstrumentedList
from sqlalchemy.orm.exc import NoResultFound
from werkzeug.exceptions import Forbidden
from werkzeug.utils import cached_property
from ...fields import ToOne
from ...backends import Pagination
from ...exceptions import ItemNotFound
from .permission import HybridPermission
from .needs import HybridItemNeed, HybridUserNeed
from ..alchemy import SQLAlchemyManager
from flask_principal import Permission, RoleNeed

PERMISSION_DEFAULTS = (
    ('read', 'yes'),
    ('create', 'no'),
    ('update', 'create'),
    ('delete', 'update')
)

DEFAULT_METHODS = ('read', 'create', 'update', 'delete')


class PrincipalsManager(SQLAlchemyManager):

    def __init__(self, resource, model):
        super(PrincipalsManager, self).__init__(resource, model)

        raw_needs = dict(PERMISSION_DEFAULTS)
        raw_needs.update(resource.meta.get('permissions', {}))
        self._raw_needs = raw_needs

    @cached_property
    def _needs(self):
        needs_map = self._raw_needs.copy()
        methods = needs_map.keys()

        def convert(method, needs, map, path=()):
            options = set()

            if isinstance(needs, six.string_types):
                needs = [needs]
            if isinstance(needs, set):
                return needs

            for need in needs:
                if need in ('no', 'nobody', 'noone'):
                    options.add(Permission(('permission-denied',)))
                elif need in ('yes', 'everybody', 'anybody', 'everyone', 'anyone'):
                    return {True}
                elif need in methods:
                    if need == method:
                        options.add(HybridItemNeed(method, self.resource))
                    elif need in path:
                        raise RuntimeError('Circular permissions in {} (path: {})'.format(self.resource, path))
                    else:
                        path += (method, )
                        options |= convert(need, map[need], map, path)

                elif ':' in need:
                    role, value = need.split(':')
                    field = self.resource.schema.fields[value]

                    if field.attribute is None:
                        field.attribute = value

                    # TODO implement this for ToMany as well as ToOne
                    if isinstance(field, ToOne):
                        target = field.target

                        if role == 'user':
                            options.add(HybridUserNeed(field))
                        elif role == 'role':
                            options.add(RoleNeed(value))
                        else:
                            for imported_need in target.manager._needs[role]:
                                if isinstance(imported_need, HybridItemNeed):
                                    imported_need = imported_need.extend(field)
                                options.add(imported_need)
                else:
                    options.add(RoleNeed(need))

            return options

        for method, needs in needs_map.items():
            converted_needs = convert(method, needs, needs_map)
            needs_map[method] = converted_needs

        return needs_map

    @cached_property
    def _permissions(self):
        permissions = {}

        for method, needs in self._needs.items():
            if True in needs:
                needs = set()
            permissions[method] = HybridPermission(*needs)

        return permissions

    def get_permissions_for_item(self, item):
        """
        Returns a dictionary of evaluated permissions for an item.

        :param item:
        :return: Dictionary in the form ``{method: bool, ..}``
        """
        return {method: permission.can(item) for method, permission in self._permissions.items()}

    def can_create_item(self, item):
        """
        Looks up permissions on whether an item may be created.

        :param item:
        """
        permission = self._permissions['create']
        return permission.can(item)

    def can_update_item(self, item, changes=None):
        """
        Looks up permissions on whether an item may be updated.

        :param item:
        :param changes: dictionary of changes
        """
        permission = self._permissions['update']
        return permission.can(item)

    def can_delete_item(self, item):
        """
        Looks up permissions on whether an item may be deleted.

        :param item:
        """
        permission = self._permissions['delete']
        return permission.can(item)

    def relation_instances(self, item, attribute, target_resource, page=None, per_page=None):
        query = getattr(item, attribute)

        if isinstance(query, InstrumentedList):
            if page and per_page:
                return Pagination.from_list(query, page, per_page)
            return query

        if isinstance(target_resource.manager, PrincipalsManager):
            read_permission = target_resource.manager._permissions['read']
            query = read_permission.apply_filters(query)

        if page and per_page:
            return query.paginate(page=page, per_page=per_page)
        return query.all()

    def paginated_instances(self, page, per_page, where=None, sort=None):
        instances = self.instances(where=where, sort=sort)
        if isinstance(instances, list):
            return Pagination.from_list(instances, page, per_page)
        return instances.paginate(page=page, per_page=per_page)

    def _query(self):
        read_permission = self._permissions['read']
        query = read_permission.apply_filters(self.model.query)

        if query is None:
            # abort with 403, but only if permissions for this resource are role-based.
            if all(need.method == 'role' for need in read_permission.needs):
                # abort(403, message='Permission denied: not allowed to access this resource')
                raise Forbidden()

        return query

    def instances(self, where=None, sort=None, page=None, per_page=None):
        """
        Applies permissions to query and returns query.

        :raises HTTPException: If read access is entirely forbidden.
        """
        query = self._query()

        if query is None:
            return []

        if where:
            query = query.filter(self._where_expression(where))
        if sort:
            query = query.order_by(*self._order_by(sort))
        return query

    def read(self, id):
        try:
            query = self._query()
            if query is None:
                raise ItemNotFound(self.resource, id=id)
            # NOTE SQLAlchemy's .get() does not work well with .filter(), therefore using .one()
            return query.filter(self.id_column == id).one()
        except NoResultFound:
            raise ItemNotFound(self.resource, id=id)

    def create(self, properties, commit=True):
        if not self.can_create_item(properties):
            raise Forbidden()
        return super(PrincipalsManager, self).create(properties, commit)

    def update(self, item, changes, *args, **kwargs):
        if not self.can_update_item(item, changes):
            raise Forbidden()
        return super(PrincipalsManager, self).update(item, changes, *args, **kwargs)

    def delete(self, item):
        if not self.can_delete_item(item):
            raise Forbidden()
        return super(PrincipalsManager, self).delete(item)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from setuptools import setup, find_packages

setup(
    name='Flask-Potion',
    version='0.1.1',
    packages=find_packages(exclude=['*tests*']),
    url='http://potion.readthedocs.org/en/latest/',
    license='MIT',
    author='Lars Schöning',
    author_email='lays@biosustain.dtu.dk',
    description='REST API framework for Flask and SQLAlchemy',
    test_suite='nose.collector',
    tests_require=[
        'Flask-Testing>=0.4.1',
        'Flask-Principal>=0.4.0',
        'nose>=1.1.2',
    ],
    install_requires=[
        'Flask>=0.10',
        'Flask-SQLAlchemy>=2.0',
        'jsonschema>=2.4.0',
        'aniso8601>=0.84',
        'blinker>=1.3',
        'six>=1.8.0'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    zip_safe=False,
    extras_require={
        'docs': ['sphinx', 'Flask-Principal'],
        'principal': [
            'Flask-Principal',
        ]
    }
)



from collections import namedtuple
from . import fields, ValidationError
import collections

from flask import json, request
from flask_sqlalchemy import Pagination as SAPagination
from werkzeug.utils import cached_property
from .exceptions import InvalidJSON
from .backends import Pagination
from .fields import ToMany
from .reference import ResourceBound
from .utils import get_value
from .schema import Schema


Comparator = namedtuple('Comparator', ['name', 'schema', 'expression', 'supported_types'])

DEFAULT_COMPARATORS = (
    Comparator('$eq',
               lambda field: field.response,
               lambda eq, value: value == eq,
               (fields.Boolean, fields.String, fields.Integer, fields.Number, fields.ToOne)),
    Comparator('$ne',
               lambda field: field.response,
               lambda ne, value: value != ne,
               (fields.Boolean, fields.String, fields.Integer, fields.Number)),
    Comparator('$in',
               lambda field: {
                   "type": "array",
                   # "minItems": 1, # NOTE: Permitting 0 items for now.
                   "uniqueItems": True,
                   "items": field.response  # NOTE: None is valid.
               },
               lambda in_, value: value in in_,
               (fields.String, fields.Integer, fields.Number)),
    Comparator('$lt',
               lambda field: {"type": "number"},
               lambda lt, value: value < lt,
               (fields.Integer, fields.Number)),
    Comparator('$gt',
               lambda field: {"type": "number"},
               lambda gt, value: value > gt,
               (fields.Integer, fields.Number)),
    Comparator('$lte',
               lambda field: {"type": "number"},
               lambda lte, value: value <= lte,
               (fields.Integer, fields.Number)),
    Comparator('$gte',
               lambda field: {"type": "number"},
               lambda gte, value: value <= gte,
               (fields.Integer, fields.Number)),
    Comparator('$text',
               lambda field: {
                   "type": "string",
                   "minLength": 1
               },
               lambda text, value: value and text in value,
               (fields.String,)),
    Comparator('$startswith',  # TODO case insensitive
               lambda field: {
                   "type": "string",
                   "minLength": 1
               },
               lambda sw, value: value and value.startswith(sw),
               (fields.String,)),
    Comparator('$endswith',  # TODO case insensitive
               lambda field: {
                   "type": "string",
                   "minLength": 1
               },
               lambda ew, value: value and value.endswith(ew),
               (fields.String,))

)

COMPARATORS = {c.name: c for c in DEFAULT_COMPARATORS}

COMPARATORS_BY_TYPE = {
    t: [c for c in DEFAULT_COMPARATORS if t in c.supported_types]
    for t in (fields.Boolean, fields.String, fields.Integer, fields.Number, fields.ToOne)
}

EQUALITY_COMPARATOR = '$eq'

ALL = '*'


class Condition(object):
    def __init__(self, attribute, comparator, value):
        self.attribute = attribute
        self.comparator = comparator
        self.value = value

    def __call__(self, item):
        return self.comparator.expression(self.value, get_value(self.attribute, item, None))


class PaginationMixin(object):
    query_params = ()

    def format_response(self, data):
        if not isinstance(data, (Pagination, SAPagination)):
            return self.format(data)

        links = [(request.path, data.page, data.per_page, 'self')]

        if data.has_prev:
            links.append((request.path, 1, data.per_page, 'first'))
            links.append((request.path, data.page - 1, data.per_page, 'prev'))
        if data.has_next:
            links.append((request.path, data.page + 1, data.per_page, 'next'))

        links.append((request.path, data.pages, data.per_page, 'last'))

        # FIXME links must contain filters & sort
        # TODO include query_params

        headers = {
            'Link': ','.join(('<{0}?page={1}&per_page={2}>; rel="{3}"'.format(*link) for link in links)),
            'X-Total-Count': data.total
        }

        return self.format(data.items), 200, headers


class RelationInstances(PaginationMixin, ToMany):
    pass


class Instances(PaginationMixin, Schema, ResourceBound):
    """
    This is what implements all of the pagination, filter, and sorting logic.

    Works like a field, but reads 'where' and 'sort' query string parameters as well as link headers.
    """
    query_params = ('where', 'sort')

    def __init__(self, reference, default_sort=None, filters=None):
        self.allowed_filters = filters
        self.filters = {}
        self.sort_fields = []

    def bind(self, resource):
        super(Instances, self).bind(resource)

        fs = resource.schema
        filters = self.allowed_filters

        # TODO only allow filters supported by the manager

        if filters in (ALL, None):
            filters = ALL
        elif isinstance(filters, (list, tuple)):
            filters = {field: ALL for field in filters}
        elif isinstance(filters, dict):
            pass

        for name, field in fs.fields.items():
            try:
                available_comparators = COMPARATORS_BY_TYPE[field.__class__]
            except KeyError:
                continue

            if filters == ALL:
                self.filters[name] = field, available_comparators
            elif name in filters:
                if filters[name] == ALL:
                    comparators = available_comparators
                else:
                    comparators = [c for c in filters[name] if c in available_comparators]

                self.filters[name] = field, comparators

        if filters in (ALL, None):
            sort = fs.fields
        elif isinstance(filters, (list, tuple, dict)):
            sort = {name: fs.fields[name] for name in filters}
        else:
            raise RuntimeError("Meta.allowed_filters is not configured properly")

        self.sort_fields = {name: field for name, field in sort.items() if self._is_sortable(field)}

    @classmethod
    def _is_sortable(cls, field):
        return isinstance(field, (fields.String, fields.Boolean, fields.Number, fields.Integer))

    def _filter_field_schema(self, field, comparators):
        if len(comparators) == 1 and comparators[0].name == EQUALITY_COMPARATOR:
            return comparators[0].schema(field)

        comparator_options = {
            "type": "object",
            "properties": {c.name: c.schema(field) for c in comparators if c.name != EQUALITY_COMPARATOR},
            "minProperties": 1,
            "maxProperties": 1,
            "additionalProperties": False
        }

        if COMPARATORS[EQUALITY_COMPARATOR] in comparators:
            return {
                "oneOf": [
                    comparator_options,
                    COMPARATORS[EQUALITY_COMPARATOR].schema(field)
                ]
            }

        return comparator_options

    @cached_property
    def _where_schema(self):
        return {
            "type": "object",
            "properties": {
                name: self._filter_field_schema(field, comparators)
                for name, (field, comparators) in self.filters.items()
            },
            "additionalProperties": False
        }

    @cached_property
    def _sort_schema(self):
        return {
            "type": "object",
            "properties": {  # FIXME switch to tuples
                             name: {"type": "boolean", "title": "Reverse order?"}
                             for name in self.sort_fields
            },
            "additionalProperties": False
        }

    def schema(self):
        request_schema = {
            "type": "object",
            "properties": {
                "where": self._where_schema,
                "sort": self._sort_schema,
                "page": {
                    "type": "integer",
                    "minimum": 1,
                    "default": 1
                },
                "per_page": {
                    "type": "integer",
                    "minimum": 1,
                    "maximum": self.resource.api.max_per_page,
                    "default": self.resource.api.default_per_page
                }
            },
            "additionalProperties": True
        }

        response_schema = {
            "type": "array",
            "items": {"$ref": "#"}
        }

        return response_schema, request_schema

    # def convert(self, value):
    # pass
    #     # TODO properties -> field attributes
    #     # parse filters

    def _convert_where(self, where):
        for name, condition in where.items():
            field, comparators = self.filters[name]

            value = None
            comparator = None
            if isinstance(condition, dict) and '$ref' not in condition:
                # if len(condition) == 1 and '$ref' in condition:

                for c in comparators:
                    if c.name in condition:
                        comparator = c
                        value = condition[c.name]
                        break

                assert comparator is not None
            elif isinstance(condition, list):
                comparator = COMPARATORS['$in']
                value = condition
            else:
                comparator = COMPARATORS['$eq']
                value = field.convert(condition)

            yield Condition(field.attribute or name, comparator, value)

    def _convert_sort(self, sort):
        for name, reverse in sort.items():
            field = self.sort_fields[name]
            yield field.attribute or name, reverse

    def parse_request(self, request):

        # TODO convert instances to FieldSet
        # TODO (implement in FieldSet too:) load values from request.args
        try:
            page = request.args.get('page', 1, type=int)
            per_page = request.args.get('per_page', self.resource.api.default_per_page, type=int)
            where = json.loads(request.args.get('where', '{}'))  # FIXME
            sort = json.loads(request.args.get('sort', '{}'), object_pairs_hook=collections.OrderedDict)
        except ValueError:
            raise InvalidJSON()

        result = self.convert({
            "page": page,
            "per_page": per_page,
            "where": where,
            "sort": sort
        })

        result['where'] = tuple(self._convert_where(result['where']))
        result['sort'] = tuple(self._convert_sort(result['sort']))
        return result

    def format(self, items):
        return [self.resource.schema.format(item) for item in items]
